/*
  png2card by CrowTRobo
	Converts a PNG image into GC memcard banner/icon data.
	September 25, 2003.
*/

#include <stdio.h>
#include <stdlib.h>
#include "png.h"

FILE *fp, *out = NULL;
png_structp png;
png_infop info, end_info;
png_bytep *row_pointers = NULL;
int w, h;

void PrintType(int type) {

	printf("Type = ");

	if (type == PNG_COLOR_TYPE_GRAY)
		printf("Grayscale\n");
  else if (type == PNG_COLOR_TYPE_GRAY_ALPHA)
		printf("Grayscale Alpha\n");
  else if (type == PNG_COLOR_TYPE_PALETTE)
		printf("Palette\n");
  else if (type == PNG_COLOR_TYPE_RGB)
		printf("RGB\n");
  else if (type == PNG_COLOR_TYPE_RGB_ALPHA)
		printf("RGB Alpha\n");
	else
		printf("Unknown\n");
}

void Error(char *str) {

	int i;

	printf("Error: %s");

	if (row_pointers != NULL) {

	  for (i = 0; i < h; i++)
		  png_free(png, row_pointers[i]);

	  png_free(png, row_pointers);
	}

	png_destroy_read_struct(&png, &info, &end_info);
	fclose(fp);

	if (out != NULL)
	  fclose(out);
}

/*
  Pack colors into 5 bits each.
*/

unsigned short RGB5(unsigned char r, unsigned char g, unsigned char b) {

	r &= 0xf8;
	g &= 0xf8;
	b &= 0xf8;

	return ((1 << 15) | (r << 7) | (g << 2) | (b >> 3));
}

/*
  Pack colors into 4 bits each with 3 bits for alpha.
*/

unsigned short RGB4A3(unsigned char r, unsigned char g, unsigned char b, unsigned char a) {

	r &= 0xf0;
	g &= 0xf0;
	b &= 0xf0;
	a &= 0xe0;

	return ((a << 7) | (r << 4) | (g) | (b >> 4));
}

/* 
  GC texture data needs to be arranged in tiles of 32 bytes.
	Textures that have 16bit pixels need to be arranged in 4x4
	tiles (2 bytes per pixel * 4 * 4 = 32 bytes).
*/

void TileRGB(int type, int chan, int x, int y) {

	png_bytep currow;
	unsigned short val;
	int i, j;

	for (i = y; i < y+4; i++) {
		currow = row_pointers[i];

		if (!(i % 2))
			fprintf(out, "\n  ");

		for (j = x; j < x+4; j++) {

			if (type == PNG_COLOR_TYPE_RGB) {
				val = RGB5(currow[j*chan], currow[j*chan+1], currow[j*chan+2]);
				fprintf(out, "0x%04X, ", val);
			}
			else {
				val = RGB4A3(currow[j*chan], currow[j*chan+1], currow[j*chan+2], currow[j*chan+3]);
				fprintf(out, "0x%04X, ", val);
			}
		}
	}
}

/* 
  GC texture data needs to be arranged in tiles of 32 bytes.
	Textures that have 8bit pixels need to be arranged in 8x4
	tiles (1 byte per pixel * 8 * 4 = 32 bytes).
*/

void TileC8(int x, int y) {

	png_bytep currow;
	int i, j;

	for (i = y; i < y+4; i++) {
		currow = row_pointers[i];
		fprintf(out, "\n  ");

		for (j = x; j < x+8; j++)
			fprintf(out, "0x%02X, ", currow[j]);
	}
}

/*
  Tiles image data.
*/

void Tile(int type, int chan) {
	
	int i, j, tilew, tileh;

	if (type == PNG_COLOR_TYPE_PALETTE) {
		tilew = 8;
		tileh = 4;
	}
	else {
		tilew = 4;
		tileh = 4;
	}

	for (i = 0; i < h; i+=tileh) {

		for (j = 0; j < w; j+=tilew) {

			if (type == PNG_COLOR_TYPE_PALETTE)
				TileC8(j, i);
			else
				TileRGB(type, chan, j, i);
		}
	}
}

int main (int argc, char *argv[]) {

	png_colorp palette;
	unsigned char header[8];
	unsigned short val;
	int i, depth, type, num_palette, channels;

	printf("\npng2card by CrowTRobo\n\n");

	if (argc < 4) {
		printf("png2card <png file> <out file> <array name>\n");
		return 1;
	}

	if ((fp = fopen(argv[1], "rb")) == NULL) {
		printf("Could not open %s for reading.\n", argv[1]);
		return 1;
	}

	if ((out = fopen(argv[2], "at")) == NULL){
		printf("Could not open %s for writing.\n", argv[2]);
		fclose(fp);
		return 1;
	}

	fread(header, 1, 8, fp);

	if (png_sig_cmp(header, 0, 8)) {
		printf("%s is not a valid PNG file.\n", argv[1]);
		fclose(fp);
		fclose(out);
		return 1;
	}

	png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

	if (png == NULL) {
		printf("Error creating read struct.\n");
		fclose(fp);
		fclose(out);
		return 1;
	}

	info = png_create_info_struct(png);

	if (info == NULL) {
		printf("Error creating info struct.\n");
		png_destroy_read_struct(&png, NULL, NULL);
		fclose(fp);
		fclose(out);
		return 1;
	}

	end_info = png_create_info_struct(png);

	if (end_info == NULL) {
		printf("Error creating end_info struct.\n");
		png_destroy_read_struct(&png, &info, NULL);
		fclose(fp);
		fclose(out);
		return 1;
	}

	png_init_io(png, fp);
	png_set_sig_bytes(png, 8);
	png_read_info(png, info);
	png_get_IHDR(png, info, (unsigned long *)&w, (unsigned long *)&h, &depth, &type, NULL, NULL, NULL);
	channels = png_get_channels(png, info);

	printf("Width = %d\n", w);
	printf("Height = %d\n", h);
	printf("Bit depth = %d\n", depth);
	PrintType(type);
	printf("Channels = %d\n", channels);

	// w & h must be 96x32 (banner) or 32x32 (icon)

	if (h != 32) {
		Error("The height is not 32.\n");
		return 1;
	}

	if (w != 96 && w != 32) {
		Error("The width is not 96 (banner) or 32 (icon).\n");
		return 1;
	}

	// now check color type, channels and bit depth. 

	if (type == PNG_COLOR_TYPE_PALETTE) {

		if (channels != 1) {
			Error("Wrong # of channels for Palette type.\n");
			return 1;
		}

		png_get_PLTE(png, info, &palette, &num_palette);

		if (num_palette > 256) {
			Error("More than 256 palette entries exist.\n");
			return 1;
		}

		printf("# palette: %d\n", num_palette);
	}
	else if (type == PNG_COLOR_TYPE_RGB) {
		
		if (channels != 3) {
			Error("Wrong # of channels for RGB type.\n");
			return 1;
		}
	}
	else if (type == PNG_COLOR_TYPE_RGB_ALPHA) {
		
		if (channels != 4) {
			Error("Wrong # of channels for Palette type.\n");
			return 1;
		}
	}
	else {
		
		if (channels != 1) {
			Error("Type unsupported.\n");
			return 1;
		}
	}

	// bit depth must be 8

	if (depth != 8) {
		Error("Bit depth is not 8.\n");
		return 1;
	}

	row_pointers = (png_bytep *)png_malloc(png, h*sizeof(png_bytep));

	for (i = 0; i < h; i++)
		row_pointers[i] = (png_bytep)png_malloc(png, w * channels * depth);

	png_read_image(png, row_pointers);

	if (type == PNG_COLOR_TYPE_PALETTE)
		fprintf(out, "unsigned char");
	else 
		fprintf(out, "unsigned short");

	fprintf(out, " %s[%d] = {\n  ", argv[3], w*h);
	Tile(type, channels);
	fprintf(out, "\n};\n");

	if (type == PNG_COLOR_TYPE_PALETTE) {
		fprintf(out, "\nunsigned short %s_tlut[%d] = {", argv[3], num_palette);
		
		for (i = 0; i < 256; i++) {	
			
			if (!(i % 8)) 
			  fprintf(out, "\n  ");

			val = RGB5(palette[i].red, palette[i].green, palette[i].blue);
			fprintf(out, "0x%04X, ", val);
		}

		fprintf(out, "\n};\n");
	}	

	fprintf(out, "\n");
	printf("\nDone\n");

	for (i = 0; i < h; i++)
		png_free(png, row_pointers[i]);

	png_free(png, row_pointers);
	png_destroy_read_struct(&png, &info, &end_info);
	fclose(fp);
	fclose(out);

	return 0;
}


